#include <thread>
#include <iostream>

#include "hpp/dds/dds.hpp"
#include "IDL/IDLs.hpp"

int main(int argc, char *argv[])
{
    //****************************** Domain participant ******************************************//
    const short int TYPE_CODE_MAX_SERIALIZED_LENGTH = 20000;
    const int32_t USER_DATA_MAX_LENGHT = 65536;

    // set factory qos
    dds::domain::qos::DomainParticipantFactoryQos factoryQos =
            dds::domain::DomainParticipant::participant_factory_qos();
    factoryQos << dds::core::policy::EntityFactory::ManuallyEnable();
    factoryQos->resource_limits.max_objects_per_thread(4096);
    dds::domain::DomainParticipant::participant_factory_qos(factoryQos);

    // set domain participant qos
    dds::domain::qos::DomainParticipantQos dpQos = dds::core::QosProvider::Default()
            .participant_qos(rti::core::builtin_profiles::qos_lib::baseline());

    rti::core::policy::DomainParticipantResourceLimits resource_limits_qos;
    resource_limits_qos.type_code_max_serialized_length(0);
    resource_limits_qos.reader_user_data_max_length(USER_DATA_MAX_LENGHT);
    resource_limits_qos.writer_user_data_max_length(USER_DATA_MAX_LENGHT);
    resource_limits_qos.type_object_max_serialized_length(TYPE_CODE_MAX_SERIALIZED_LENGTH);
    dpQos << resource_limits_qos;

    std::map<std::string,std::string> dpPropertyMap = {{"dds.transport.UDPv4.builtin.recv_socket_buffer_size", "2100000000"},
                                                       {"dds.transport.UDPv4.builtin.send_socket_buffer_size", "2100000000"},
                                                       {"dds.transport.UDPv4.builtin.parent.message_size_max", "65530"},
                                                       {"dds.transport.shmem.builtin.receive_buffer_size", "204857600"}};

    dpQos << rti::core::policy::Property(dpPropertyMap.begin(),dpPropertyMap.end());

    // Create a DomainParticipant
    dds::domain::DomainParticipant domainParticipant = dds::domain::DomainParticipant(1,dpQos);
    domainParticipant.enable();

    //******************************    create publisher      ******************************************//

    // choose a topicName
    std::string topicName = "MY_IDL";

    // create topic with the domainParticipant, the topicName, the DataType
    dds::topic::Topic<struct3> topic(domainParticipant,topicName);

    auto qosPub =  domainParticipant.default_publisher_qos()
            << dds::core::policy::Partition("defaultPartition");

    // create publisher with created domainParticipant
    dds::pub::Publisher publisher(domainParticipant, qosPub);

    //******************************    create data writer      ******************************************//

    std::map<std::string, std::string> qosDwPropertyMap = {{"dds.builtin_type.*.max_size", "2100000000"},
                                                           {"dds.builtin_type.*.alloc_size", "2100000000"},
                                                           {"dds.data_writer.history.memory_manager.fast_pool.pool_buffer_max_size", "700000000"}};
    dds::pub::qos::DataWriterQos qosDw;
    qosDw << rti::core::policy::Property(qosDwPropertyMap.begin(), qosDwPropertyMap.end());

    qosDw << rti::core::policy::PublishMode::Asynchronous()
          << dds::core::policy::ResourceLimits()
          << dds::core::policy::Lifespan(dds::core::Duration(10,0))
          << dds::core::policy::History::KeepAll();

    // create DataWriter opject with the publisher, the topic, and the dataWriterQos
    dds::pub::DataWriter<struct3> dataWriter(publisher, topic, qosDw);

    // wait for connecting to TEXT_RECEIVER (if dont, samples will not receive)
    std::this_thread::sleep_for(std::chrono::seconds(5));


    //******************************    fill sample     ******************************************//

    // create an object with the data type for sending sample
    struct3 sample;

    sample.num1(1000);
    sample.enum1(TestEnum::value1);
    sample.enum2(TestEnum2::value1);

    for(int i = 0 ; i < 1111 ; i++)
    {
        struct2 st2;
        st2.num1(1000);
        st2.num2(1000);
        st2.enum1(TestEnum::value1);
        st2.enum2(TestEnum::value1);
        st2.enum3(TestEnum2::value1);
        st2.enum4(TestEnum2::value1);
        for(int j = 0 ; j < 222 ; j++)
        {
            struct1 st1;
            st1.str1("KILIKILIKILIKILI");
            st1.str2("KILIKILIKILIKILI");
            st1.str3("KILIKILIKILIKILI");
            st1.enum1(TestEnum::value1);
            st1.enum2(TestEnum2::value1);
            st2.structs1().push_back(st1);
        }
        sample.structs1().push_back(st2);
    }

    //******************************    publish sample     ******************************************//
    for(int i=0; i < 5; i++)
    {
        // send the object
        dataWriter.write(sample);
        std::this_thread::sleep_for(std::chrono::seconds(10));
    }
}
